import base64
import hashlib
import json
import re

from .. import json_util


def get_str_hash(s, consider_whitespace=False):
    if not consider_whitespace:
        # \s for "any whitespace"
        s = s.replace('\\n', ' ')
        s = re.sub('\s+', ' ', s)
        s = re.sub('\s+', ' ', s)
        s = s.replace('" ', '"').replace(' "', '"')
    h = hashlib.md5()
    h.update(s.encode('utf8'))
    str_hash = h.hexdigest()
    return str_hash


def get_dict_hash(d):
    if d is None:
        d = {}
    s = json_util.dict_to_sorted(d)
    str_val = json.dumps(s, cls=json_util.CustomEncoder)
    str_hash = get_str_hash(str_val)
    return str_hash



import unittest

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from . import *


class TestSettings(unittest.TestCase):
    def test_merge(self):
        pass

    def test_tracking(self):
        pass


class TestHCSObject(unittest.TestCase):
    def test_constructor(self):
        test_obj = HCSObject()


# class TestHCS(unittest.TestCase):
#     def load_yaml(self, rel_path):
#         full_path = os.path.join(os.path.dirname(__file__), rel_path)
#         object_dict = None
#         with open(full_path, 'r') as f:
#             raw_txt = f.read()
#             object_dict = yaml.safe_load(raw_txt)
#         return object_dict

#     def test_inheritance(self):
#         type_map = {
#             'constant': distribution.Constant,
#             'uniform': distribution.Uniform,
#             'logistic': distribution.Logistic,
#             'normal': distribution.Normal,
#             'skew_normal': distribution.SkewNormal,
#             'log_normal': distribution.LogNormal,
#             'beta': distribution.Beta,
#             'gamma': distribution.Gamma,
#         }
#         object_dict = self.load_yaml('test_data/example_dist_obj.yml')

#         hcs_api = HCS(type_map, object_dict)

#         obj2 = hcs_api.get_object_hcs(2)
#         self.assertAlmostEqual(obj2.config['loc'], 10)
#         self.assertAlmostEqual(obj2.config['scale'], 2)

#         # TODO: Test inheritance class mismatches?
#         # Or... allow class mismatches (like for projects)

#     def test_cache(self):
#         pass

#     def test_draft_vs_published(self):
#         pass

#     def test_save(self):
#         pass


import unittest

import os
import json

from . import *

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
#logging.getLogger('keyrock_core.config_loader.config_loader').setLevel(logging.DEBUG)


class TestConfigUtil(unittest.TestCase):
    def test_root(self):
        ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), '.'))
        
        config = load(os.path.join(ROOT_DIR, 'test_data/example_root.yml'))
        logger.debug(json.dumps(config, indent=2))

        self.assertEquals(config['test_dict']['int_item'], 69)
        self.assertEquals(config['test_dict']['imported_item'], 'from_yaml')
        self.assertEquals(config['test_dict']['dict_item']['imported_item'], 123)
        self.assertEquals(config['test_dict']['js_item'], 'from_json')

        #config = load(os.path.join(ROOT_DIR, 'test_data/bad_example.yml'))
        #print(json.dumps(config, indent=2))



import logging
logger = logging.getLogger(__name__)

import re


def repl_last(s, sub, repl):
    pattern = sub + '(?!.*' + sub + ')'
    result = re.sub(pattern, repl, s, flags=re.DOTALL)
    return result


def strip_limit(query_str):
    if 'limit' not in query_str.lower():
        return query_str

    sub = '\s*LIMIT\s+.*'
    return repl_last(query_str, sub, '')


def set_limit(query_str, limit):
    query_str = strip_limit(query_str)
    query_str += f' LIMIT {limit}'
    return query_str


import unittest
import json

from . import *

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class TestJsonUtil(unittest.TestCase):
    def test_diff(self):
        diff_simple_eq = diff(1, 1)
        self.assertEqual(diff_simple_eq['op'], None)
        self.assertTrue('val' in diff_simple_eq)

        diff_simple_neq = diff(1, 2)
        self.assertTrue(diff_simple_neq['val'], 2)
        self.assertEqual(diff_simple_neq['op'], 'mod')

        diff_dict0 = diff({'x': 0}, {'x': 0})
        self.assertEqual(diff_dict0['op'], None)

        diff_dict1 = diff({'x': 0}, {'x': 1})
        self.assertTrue(diff_dict1['diff'])
        self.assertTrue(diff_dict1['op'], 'dict')

        diff_dict2 = diff(
            {
                'x': {'s': 0}, 
                'y': {'s': 0}
            },
            {
                'x': {'s': 0},
                'y': {'s': 1}
            }
        )
        logger.debug(json.dumps(diff_dict2, indent=2))

        self.assertEqual(diff_dict2['diff']['y']['diff']['s']['val_from'], 0)
        self.assertEqual(diff_dict2['diff']['y']['diff']['s']['val'], 1)
        self.assertEqual(diff_dict2['op'], 'dict')
        self.assertEqual(diff_dict2['diff']['y']['diff']['s']['op'], 'mod')

        # TODO: test explicit_group_list



import logging
logger = logging.getLogger(__name__)

from . import exc


def int_none(val):
    try:
        return int(val)
    except:
        return None


def local_to_tracked(obj_id, settings_local, depth=0):
    settings_tracked = {
        'd': depth,     # Depth
        'r': obj_id,    # Root/Origin Id (where it first appeared)
        'p': obj_id,    # Previous Id (where it was previously set)
        'c': obj_id,    # Current Id (where it is currently set)
        # Optional:
        # 'v': Effective Value (if leaf)
        # 'pv': Previous Value (if leaf)
        # 'dict': Dict Value (if not leaf)
        # 'list': List Value (if not leaf)
    }

    if isinstance(settings_local, dict):
        setting_dict = {}
        for key, val in settings_local.items():
            setting_dict[key] = local_to_tracked(obj_id, val, depth)
        settings_tracked['dict'] = setting_dict
    elif isinstance(settings_local, list):
        setting_list = []
        for val in settings_local:
            setting_list.append(local_to_tracked(obj_id, val, depth))
        settings_tracked['list'] = setting_list
    else:
        settings_tracked['v'] = settings_local  # Effective Value
        settings_tracked['pv'] = settings_local # Parent Value

    return settings_tracked


def tracked_to_local(settings_tracked):
    if settings_tracked is None:
        return None

    if 'v' in settings_tracked:
        return settings_tracked['v']
    elif 'dict' in settings_tracked:
        settings_local = {}
        for key, val_tracked in settings_tracked['dict'].items():
            settings_local[key] = tracked_to_local(val_tracked)
        return settings_local
    elif 'list' in settings_tracked:
        settings_local = []
        for val_tracked in settings_tracked['list']:
            settings_local.append(tracked_to_local(val_tracked))
        return settings_local
    else:
        raise exc.InvalidTrackedSetting(settings_tracked)


def merge_tracked_settings(settings_parent, settings_child):
    '''
    Include all the parent settings
      merge with child settings if they exist
    Include the child settings which don't already exist in the parent
      theoretically could return settings_merged from every branch
    '''

    if settings_child is None:
        settings_child = {}

    settings_merged = {
        'r': settings_parent['r'],
    }

    # Simple value settings
    if 'v' in settings_parent:
        if 'v' in settings_child:
            settings_merged.update({
                'c': settings_child['c'],   # Child Id
                'd': settings_child['d'],   # Depth
                'p': settings_parent['c'],  # Parent Id
                'pv': settings_parent['v'], # Parent Value
                'v': settings_child['v'],   # Child Value
            })
        else:
            settings_merged.update({
                'c': settings_parent['c'],   # Child Id
                'd': settings_parent['d'],   # Depth
                'p': settings_parent['c'],   # Parent Id
                'pv': settings_parent['v'],  # Parent Value
                'v': settings_parent['v'],   # Child Value
            })
    elif 'v' in settings_child:
        settings_merged = settings_child
    else:
        # Handle dictionary (still record parent and child ids, and origin depth)
        settings_merged.update({
            'p': settings_parent['c'],                          # Parent Id
            'c': settings_child.get('c', settings_parent['c']), # Child Id
            'd': settings_child.get('d', settings_parent['d']), # Origin depth
        })

    # Dict settings
    if 'dict' in settings_parent:
        # Merge the tracked data from each dict entry
        child_dict = settings_child.get('dict', {})
        merged_dict = {}
        for key, parent_entry in settings_parent['dict'].items():
            child_entry = child_dict.get(key)
            merged_dict[key] = merge_tracked_settings(
                parent_entry,
                child_entry
            )
        settings_merged['dict'] = merged_dict
    # In case only the child has this dict
    if 'dict' in settings_child:
        merged_dict = settings_merged.setdefault('dict', {})
        for key, child_entry in settings_child['dict'].items():
            if key not in merged_dict:
                merged_dict[key] = child_entry

    # List settings
    if 'list' in settings_parent:
        settings_merged['list'] = settings_parent['list']
    if 'list' in settings_child:
        settings_merged.setdefault('list', []).extend(settings_child['list'])

    return settings_merged



import json
import marshmallow as m
import logging
logger = logging.getLogger(__name__)

from .. import json_util

from .schema import TypeConfigFactory
from . import datastore
from . import settings
from . import exc


class HCS():
    class BaseSchema(m.Schema):
        class Meta:
            include = {
                'title': m.fields.String(required=True),
                'object_class': m.fields.String(required=True),
                'project_id': m.fields.Integer(missing=None),
                'parent_id': m.fields.Integer(missing=None),
            }

    class StoreSchema(BaseSchema):
        class Meta:
            include = {
                'settings_draft': m.fields.Dict(missing=None),
                'settings_published': m.fields.Dict(missing=None),
                'version_draft': m.fields.Integer(missing=0),
                'version_published': m.fields.Integer(missing=0),
            }

    class CacheSchema(BaseSchema):
        class Meta:
            include = {
                'settings': m.fields.Dict(required=True),
                'version': m.fields.Integer(required=True),
            }

    class EditorSchema(StoreSchema):
        class Meta:
            include = {
                'settings_tracked': m.fields.Dict(missing=None),
            }

    def __init__(self, type_map, src):
        # Factory for creating objects
        self.obj_factory = TypeConfigFactory(type_map)

        # Datastore for accessing raw obj data
        self.datastore = None
        self.cache = None
        if isinstance(src, dict):
            self.datastore = datastore.DictDatastore(src)
            self.cache = datastore.DictDatastore({})
        else:
            # TODO: Connection string for SQLAlchemy
            assert(False)

        # TODO: Have to clear the cache when the server comes online
        # or after recompile,
        # But don't want to do this every time an HCS object is created
        #self.cache.clear()

    def get_object_hcs(self, obj_id, default_type=None, published=True):
        if obj_id is None and default_type is None:
            raise Exception('type required for new object')

        object_merged = self.get_object_merged(obj_id, default_type=default_type, published=published)
        hcs_obj = self.obj_factory.get_instance(object_merged['object_class'], object_merged['settings'])
        hcs_obj.set_meta(obj_id=obj_id, title=object_merged['title'])

        #print('HCS OBJECT', hcs_obj)
        return hcs_obj

    def get_object_merged(self, obj_id, default_type=None, published=True):
        if not published:
            # If not published, have to build the merged object, but not save it
            object_merged = self.make_cache_object(obj_id, default_type, False)
            return object_merged

        object_merged = self.cache.get(obj_id, self.CacheSchema())
        if object_merged is None:
            # Build the merged object and save it
            object_merged = self.make_cache_object(obj_id, default_type, True)
            self.cache.set(obj_id, object_merged, self.CacheSchema())

        #print('MERGED OBJECT', json.dumps(object_merged, indent=2))
        return object_merged

    def make_cache_object(self, obj_id, default_type, published=True):
        object_raw = self.get_object_raw(obj_id, default_type=default_type)
        object_class = object_raw['object_class']

        if published:
            version = object_raw['version_published']
        else:
            version = object_raw['version_draft']

        settings_merged = self.get_settings_merged(obj_id, default_type=object_class)
        object_merged = HCS.CacheSchema().load({
            'object_class': object_class,
            'title': object_raw['title'],
            'project_id': object_raw['project_id'],
            'parent_id': object_raw['parent_id'],
            'settings': settings_merged,
            'version': version,
        })
        # TODO:
        # 2. Resolve ids to titles (project_id, parent_id) -> project_name
        # 3. Post process / cache object-specific transient data (dependency stuff?)
        return object_merged

    def get_object_raw(self, obj_id, default_type=None):
        if obj_id is None:
            # Fill out a new object
            # Use settings from default_type schema
            hcs_class = self.obj_factory.get_class(default_type)
            settings_default = hcs_class.ConfigSchema().load({})
            #print('DEFAULT SETTINGS', settings_default)
            obj_raw = self.StoreSchema().load({
                'title': 'Untitled',
                'object_class': default_type,
                'settings_published': settings_default,
            })
        else:
            obj_raw = self.datastore.get(obj_id, self.StoreSchema())
            if obj_raw is None:
                raise exc.ObjectNotFoundError(obj_id)

        #print('RAW OBJECT', obj_id, json.dumps(obj_raw, indent=2))
        return obj_raw

    def get_settings_merged(self, obj_id, default_type, published=True):
        settings_tracked = self.get_settings_tracked(obj_id, default_type=default_type, published=published)
        settings_merged = settings.tracked_to_local(settings_tracked)
        if settings_merged is None:
            settings_merged = {}
        return settings_merged

    def get_settings_tracked(self, obj_id, default_type, published=True, recursion_list=None):
        # Create the recursion list if it doesn't exist yet
        if recursion_list is None:
            recursion_list = []
        depth = len(recursion_list)

        local_obj = self.get_object_raw(obj_id, default_type)
        if local_obj['object_class'] != default_type:
            # This maybe doesn't need to be an error, could allow project-inheritance
            raise Exception("object_class mismatch: found '{0}' but expected '{1}'")

        if published:
            local_settings = local_obj['settings_published']
        else:
            local_settings = local_obj['settings_draft']

        if local_settings is None:
            local_settings = {}

        local_settings_tracked = settings.local_to_tracked(obj_id, local_settings, depth)

        parent_id = local_obj['parent_id']
        if obj_id in recursion_list:
            # Break the recursion
            parent_id = None
        else:
            recursion_list.append(obj_id)

        if obj_id is None and parent_id is None:
            # This is a default/root object and has no parent
            return local_settings_tracked

        parent_settings_tracked = self.get_settings_tracked(parent_id, default_type=default_type, published=True)
        merged_settings_tracked = settings.merge_tracked_settings(parent_settings_tracked, local_settings_tracked)
        return merged_settings_tracked



import os
import yaml
import json

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def load(filepath):
    logger.debug('Load config: {}'.format(filepath))

    root_folder = os.path.dirname(filepath)
    filename = os.path.basename(filepath)

    config_dict = _load_file(root_folder, filename)
    config_dict = _resolve_inheritance(config_dict, config_dict)

    return config_dict


def _load_file(dirname, filename):
    logger.debug('load file: {} {}'.format(dirname, filename))

    file_ext = os.path.splitext(filename)[1]
    filepath = os.path.join(dirname, filename)

    final_dict = {}

    if file_ext in ['.yml', '.yaml']:
        config_dict = _load_yml(filepath)
    elif file_ext in ['.js', '.json']:
        config_dict = _load_json(filepath)
    else:
        logger.error('unrecognized file type: {}'.format(filename))
        return None

    # Assign dot-separated keys to the appropriate nodes
    config_dict = _resolve_dot_values(config_dict)

    if config_dict is not None:
        import_list = config_dict.get('import', [])
        for import_filename in import_list:
            logger.debug('import: {}'.format(import_filename))
            import_dict = _load_file(dirname, import_filename)
            _merge_config(final_dict, import_dict)

    _merge_config(final_dict, config_dict)

    return final_dict


def _load_yml(filepath):
    try:
        with open(filepath) as file:
            file_text = file.read()
            return yaml.load(file_text, Loader=yaml.SafeLoader)
    except Exception as e:
        logger.error(e)
        return {}


def _load_json(filepath):
    try:
        with open(filepath) as file:
            file_text = file.read()
            return json.loads(file_text)
    except Exception as e:
        logger.error(e)
        return {}


def _merge_config(final_dict, layer_dict):
    if layer_dict is None:
        return

    for key, layer_val in layer_dict.items():
        if key != 'import':
            current_val = final_dict.get(key)
            if current_val is None:
                final_dict[key] = layer_val
            elif isinstance(current_val, dict):
                if isinstance(layer_val, dict):
                    _merge_config(current_val, layer_val)
                else:
                    logger.warning('layer value overwrites dict')
                    final_dict[key] = layer_val
            else:
                final_dict[key] = layer_val


def _resolve_inheritance(root_node, node):
    if node is None:
        return None

    if isinstance(node, dict):
        final_node = {}
        if 'inherit' in node:
            # Load inherited base values first
            for abs_path in node['inherit']:
                inherit_node = _find_node(root_node, abs_path)
                _merge_config(final_node, inherit_node)

            # Copy the remaining values
            for key, val in node.items():
                if key != 'inherit':
                    final_node[key] = _resolve_inheritance(root_node, val)

            return final_node

    if isinstance(node, list):
        final_node = []
        for val in node:
            final_node.append(_resolve_inheritance(root_node, val))

        return final_node

    return node


def _resolve_dot_values(node):
    if node is None:
        return None

    if isinstance(node, dict):
        final_node = {}
        for key, val in node.items():
            final_val = _resolve_dot_values(val)
            _assign_val_to_node(final_node, key, final_val)
        return final_node
    elif isinstance(node, list):
        final_node = []
        for val in node:
            final_node.append(_resolve_dot_values(val))
        return final_node
    else:
        return node


def _find_node(search_node, abs_path):
    keys = abs_path.split('.')
    node = search_node
    for key in keys:
        node = node.get(key)
        if node is None:
            return None
    return node


def _assign_val_to_node(search_node, rel_path, val):
    node = search_node

    keys = rel_path.split('.')
    last_key = keys.pop()
    for key in keys:
        node = node.setdefault(key, {})
        if not isinstance(node, dict):
            raise Exception(f'node already has non-dict value: {key} of {rel_path}')

    if last_key not in node:
        # Nothing to merge
        node[last_key] = val
    else:
        current_val = node[last_key]
        if isinstance(current_val, dict):
            _merge_config(current_val, val)
        elif isinstance(current_val, list):
            raise NotImplementedError()
        else:
            node[last_key] = val



import collections
from dataclasses import make_dataclass
import marshmallow as m


class DataClassSchema(m.Schema):
    @m.post_load
    def create_object(self, dict_val, **kwargs):
        class_name = type(self).__name__ + '_dataclass'
        obj = make_dataclass(class_name, [(k, type(v)) for k,v in dict_val.items()])
        for k, v in dict_val.items():
            setattr(obj, k, v)
        return obj


class TupleSchema(m.Schema):
    @m.post_load
    def create_namedtuple(self, dict_val, **kwargs):
        class_name = type(self).__name__ + '_tuple'
        keys = dict_val.keys()
        values = dict_val.values()
        return collections.namedtuple(class_name, keys)(*values)


class TypeConfigSchema(m.Schema):
    class Meta:
        include = {
            'type': m.fields.Str(required=True),
            'config': m.fields.Dict(missing={}),
        }


class TypeConfigFactory():
    def __init__(self, type_map):
        self.type_map = type_map

    def get_class(self, obj_type):
        inst_class = self.type_map.get(obj_type)
        if inst_class is None:
            raise Exception('type not found: {}'.format(type_config['type']))
        return inst_class

    def get_instance(self, obj_type, obj_config, **kwargs):
        inst_class = self.get_class(obj_type)
        return inst_class(obj_config, **kwargs)

    def from_type_config(self, type_config, **kwargs):
        if isinstance(type_config, str):
            type_config = {'type': type_config}

        inst_class = self.type_map.get(type_config['type'])
        if inst_class is None:
            raise Exception('type not found: {}'.format(type_config['type']))
        return inst_class(type_config.get('config', {}), **kwargs)

    def from_type_config_list(self, type_config_list, **kwargs):
        instance_list = []
        for type_config in type_config_list:
            instance_list.append(self.from_type_config(type_config, **kwargs))
        return instance_list



import logging
logger = logging.getLogger(__name__)

import json
import collections


def load_safe(json_str):
    try:
        return json.loads(json_str)
    except:
        return json_str


def dict_to_sorted(d):
    if d is None:
        return None

    if not isinstance(d, dict):
        return d

    s = collections.OrderedDict(sorted(d.items()))
    for key, val in s.items():
        if isinstance(val, dict):
            s[key] = dict_to_sorted(val)

    return s


def only_update_existing(base_dict, new_dict):
    for key, val in base_dict.items():
        # If key isn't in new_dict, keep the old val
        base_dict[key] = new_dict.get(key, val)
    return base_dict


def delta(dict_from, dict_to):
    patch = {}

    for key, val in dict_to.items():
        if key not in dict_from:
            patch[key] = val
        elif isinstance(val, dict):
            patch[key] = delta(val, dict_from[key])
        elif isinstance(val, list):
            patch[key] = val
        elif val != dict_from[key]:
            patch[key] = val

    return patch


def get_node(d, path_str, default_val=None):
    path_array = path_str.split('.')
    node = d
    for branch in path_array:
        if isinstance(node, dict):
            node = node.get(branch)
        elif isinstance(node, list):
            node = node[int(branch)]
        else:
            return default_val
    return node


def get_path_value(d, path_str, default_val=None):
    return get_node(d, path_str, default_val)


def set_path_value(node, path_str, val):
    #print('SET PATH VALUE', node['id'], path_str, val)
    path_array = path_str.split('.')
    last_branch = path_array.pop()
    for branch in path_array:
        if isinstance(node, dict):
            node = node.setdefault(branch, {})
        elif isinstance(node, list):
            node = node[int(branch)]
        else:
            raise Exception('set_path_value: existing search node is not a dict or list')

    if isinstance(node, dict):
        node[last_branch] = val
    elif isinstance(node, list):
        node[int(last_branch)] = val
    else:
        raise Exception('set_path_value: penultimate node is not a dict or list')

    return node


def prune_sub_dicts(node):
    if node is None or not isinstance(node, dict):
        return node
    for key, val in node.items():
        node[key] = omit_dicts(val)


def omit_dicts(node):
    if node is None or not isinstance(node, dict):
        return node
    pruned_node = {}
    for key, val in node.items():
        if isinstance(val, dict):
            pruned_node[key] = '...'
        else:
            pruned_node[key] = val
    return pruned_node


def find_paths_with_value(search_val, src_data, key_filter_list=None, current_path=None, depth=0, current_key=None):
    # Find all paths with this value as the RHS
    # If key_filter_list is present, only include keys from the list
    if not src_data:
        return []

    max_depth = 10
    if depth > max_depth:
        logger.warning('find_paths_with_value: max depth reached:', max_depth)
        return []
    depth += 1

    if current_path is None:
        current_path = ''

    if current_key is None:
        current_key = current_path

    if isinstance(src_data, dict):
        path_list = []
        for key, val in src_data.items():
            sub_path = '{0}.{1}'.format(current_path, key)
            sub_path_list = find_paths_with_value(
                search_val, src_data=val, key_filter_list=key_filter_list, current_path=sub_path, depth=depth, current_key=key)
            path_list.extend(sub_path_list)
        return path_list
    elif isinstance(src_data, list):
        path_list = []
        for i, val in enumerate(src_data):
            sub_path = '{0}.{1}'.format(current_path, i)
            sub_path_list = find_paths_with_value(
                search_val, src_data=val, key_filter_list=key_filter_list, current_path=sub_path, depth=depth, current_key=i)
            path_list.extend(sub_path_list)
        return path_list
    elif src_data == search_val:
        if key_filter_list is not None and current_key in key_filter_list:
            return [current_path]

    return []


def find_paths_with_key(search_key, src_data, current_path=None, depth=0):
    if not src_data:
        return []

    max_depth = 10
    if depth > max_depth:
        logger.warning('find_paths_with_key: max depth reached:', max_depth)
        return []
    depth += 1

    path_list = []
    if isinstance(src_data, dict):
        for key, val in src_data.items():
            if current_path is None:
                sub_path = key
            else:
                sub_path = '{0}.{1}'.format(current_path, key)

            if key == search_key:
                path_list.append(sub_path)
            # Continue searching recusrively for nested keys
            sub_path_list = find_paths_with_key(
                search_key, src_data=val, current_path=sub_path, depth=depth)
            path_list.extend(sub_path_list)
    elif isinstance(src_data, list):
        for i, val in enumerate(src_data):
            if current_path is None:
                sub_path = key
            else:
                sub_path = '{0}.{1}'.format(current_path, i)

            # Continue searching recusrively for nested keys
            sub_path_list = find_paths_with_key(
                search_key, src_data=val, current_path=sub_path, depth=depth)
            path_list.extend(sub_path_list)
    return path_list



import json

import logging
logger = logging.getLogger(__name__)


def diff(obj_from, obj_to, explicit_group_list=None, ignore_group_list=None, group_key=None):
    ''' explicit_group_list:
            if None:
                always delete missing items
            if not None:
                only delete items missing from groups specified in the list (by group key)
                otherwise consider them as unchanged
    '''
    if isinstance(obj_from, dict) and isinstance(obj_to, dict):
        return diff_dict(obj_from, obj_to, explicit_group_list, ignore_group_list, group_key)

    if isinstance(obj_from, list) and isinstance(obj_to, list):
        return diff_list(obj_from, obj_to, explicit_group_list, ignore_group_list, group_key)

    if str(obj_from) != str(obj_to):
        return {
            'op': 'mod',
            'val_from': obj_from,
            'val': obj_to,
        }

    return {
        'op': None,
        'val': obj_from,
    }


def diff_dict(dict_from, dict_to, explicit_group_list, ignore_group_list, group_key):
    result = {
        'op': 'dict',
        'diff': {},
    }

    has_sub_diff = False

    # Find the additions
    for key, val in dict_to.items():
        if ignore_group_list is not None and key in ignore_group_list:
            pass
        elif key not in dict_from:
            has_sub_diff = True
            result['diff'][key] = {
                'op': 'add',
                'val': val,
            }

    # Find the subtractions
    if explicit_group_list is None or group_key in explicit_group_list:
        for key, val in dict_from.items():
            if ignore_group_list is not None and key in ignore_group_list:
                pass
            elif key not in dict_to:
                has_sub_diff = True
                result['diff'][key] = {
                    'op': 'del',
                    'val_from': val,
                }

    # Find the changes
    for key, val_from in dict_from.items():
        if ignore_group_list is not None and key in ignore_group_list:
            pass
        elif key in dict_to:
            val_to = dict_to[key]
            sub_result = diff(val_from, val_to, explicit_group_list, ignore_group_list, key)
            result['diff'][key] = sub_result
            if sub_result['op'] is not None:
                has_sub_diff = True

    if not has_sub_diff:
        return {
            'op': None,
            'val': dict_from,
        }

    return result


def diff_list(list_from, list_to, explicit_group_list, ignore_group_list, group_key):
    mod_result = {
        'op': 'mod',
        'val_from': list_from,
        'val': list_to,
    }

    if len(list_from) != len(list_to):
        return mod_result

    for v in list_from:
        assert(not isinstance(v, dict)), 'dict elements not allowed in list'
        assert(not isinstance(v, list)), 'list elements not allowed in list'
        if v not in list_to:
            return mod_result

    for v in list_to:
        assert(not isinstance(v, dict)), 'dict elements not allowed in list'
        assert(not isinstance(v, list)), 'list elements not allowed in list'
        if v not in list_from:
            return mod_result

    return {
        'op': None,
        'val': list_from,
    }


def strip_unchanged(diff_node):
    op = diff_node['op']
    if op == 'dict':
        return_val = {'op': 'dict', 'dict': {}}
        for item_key, item_node in diff_node['diff'].items():
            if item_node['op'] is None:
                pass
            elif item_node['op'] == 'dict':
                return_val['dict'][item_key] = strip_unchanged(item_node)
            else:
                return_val['dict'][item_key] = item_node
        return return_val
    elif op is None:
        return None
    else:
        return diff_node


def assemble_node_data(diff_node):
    op = diff_node['op']
    if op == 'dict':
        pass
    elif op == 'del':
        assert(False), "cannot delete self"
    else:
        return diff_node['val']


def get_property_op(diff_node, prop_key):
    if prop_key in diff_node:
        op = diff_node[prop_key].get('op')
        return op
    return None


class ObjectNotFoundError(Exception):
    def __init__(self, obj_id):
        msg = 'Object not found: {0}'.format(obj_id)
        super().__init__(msg)

class AccessDenied(Exception):
    def __init__(self, obj_id):
        msg = 'Operation prohibited on: {0}'.format(obj_id)
        super().__init__(msg)

class ObjectPendingDeletion(Exception):
    def __init__(self, obj_id):
        msg = 'Object pending deletion: {0}'.format(obj_id)
        super().__init__(msg)

class InvalidTrackedSetting(Exception):
    def __init__(self, tracked_setting):
        msg = repr(tracked_setting)
        super().__init__(msg)

class VersionConflictError(Exception):
    def __init(self, old_ver, new_ver):
        msg = 'Modified version mismatch: {0}:{1}'.format(old_ver, new_ver)
        super().__init__(msg)



import json
import datetime
import decimal
import typing
import io
import math

import logging
logger = logging.getLogger(__name__)

# class CustomEncoder(json.JSONEncoder):
#     def default(self, obj):
#         if isinstance(obj, datetime.datetime):
#             return str(obj)
#         elif isinstance(obj, datetime.date):
#             return str(obj)
#         elif isinstance(obj, bytes):
#             return obj.decode("utf-8", "replace")
#         return json.JSONEncoder.default(self, obj)

class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return str(obj)
        elif isinstance(obj, datetime.date):
            return str(obj)
        elif isinstance(obj, decimal.Decimal):
            return (str(o) for o in [obj])
        elif isinstance(obj, typing.Generator):
            # @hack: deal with aggregate-function Snowflake result data
            list_conv = [o for o in obj]
            if len(list_conv) == 1:
                return list_conv[0]
            else:
                return json.dumps(list_conv)
        elif isinstance(obj, bytes):
            return obj.decode("utf-8", "replace")
        elif isinstance(obj, memoryview):
            #return obj.tobytes().decode("utf-8", "replace")
            return '[BINARY DATA]'
        return json.JSONEncoder.default(self, obj)
        #return json.dumps(obj, allow_nan=False)



import json
import marshmallow as m

from . import schema
from . import exc


class Datastore():
    def __init__(self):
        pass

    def get(self, obj_id, schema):
        raise NotImplementedError()

    def set(self, obj_id, raw_data, schema):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()


class DictDatastore(Datastore):
    def __init__(self, obj_dict):
        super().__init__()
        self.obj_dict = obj_dict

    def get(self, obj_id, schema):
        raw_data = self.obj_dict.get(obj_id)
        if raw_data is None:
            return None
        return schema.load(raw_data)

    def set(self, obj_id, raw_data, schema):
        schema.validate(raw_data)
        if obj_id in self.obj_dict:
            self.obj_dict[obj_id].update(raw_data)
        else:
            self.obj_dict[obj_id] = raw_data

    def clear(self):
        self.obj_dict = {}


class PgDatastore(Datastore):
    # Maybe use SqlAlchemy to create the schema?
    def get(self, obj_id, schema):
        raise NotImplementedError()

    def set(self, obj_id, raw_data, schema):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()



import unittest

import logging
logger = logging.getLogger(__name__)

from . import *


class TestUtil(unittest.TestCase):
    def test_set_limit_0(self):
        self.assertEqual(
            set_limit('SELECT * FROM test LIMIT 1', 1),
            'SELECT * FROM test LIMIT 1'
        )

    def test_set_limit_1(self):
        self.assertEqual(
            set_limit('SELECT * FROM test', 1),
            'SELECT * FROM test LIMIT 1'
        )

    def test_set_limit_2(self):
        self.assertEqual(
            set_limit('SELECT * FROM test LIMIT 10', 2),
            'SELECT * FROM test LIMIT 2'
        )

    def test_set_limit_3(self):
        self.assertEqual(
            set_limit('SELECT * FROM test LIMIT 10, 2', 3),
            'SELECT * FROM test LIMIT 3'
        )

    def test_set_limit_4(self):
        self.assertEqual(
            set_limit('SELECT * FROM test \n\tLIMIT  \t 10, 2', 4),
            'SELECT * FROM test LIMIT 4'
        )


import marshmallow as m
import logging
logger = logging.getLogger(__name__)

from .. import hash_util


class HCSObject():
    class ConfigSchema(m.Schema):
        class Meta:
            include = {
                'is_hcs': m.fields.Boolean(missing=True),
                'empty_hack': m.fields.Boolean(required=False),
            }

    def __init__(self, config=None):
        if config is None:
            config = {}

        self.obj_id = None
        self.title = None

        # Validate the schema and load missing values
        self.config = self.ConfigSchema().load(config)
        self.config_hash = hash_util.get_dict_hash(self.config)
        self.load_config(self.config)

    def set_meta(self, obj_id=None, title=None):
        if obj_id is not None:
            self.obj_id = obj_id
        if title is not None:
            self.title = title

    def load_config(self, config):
        # Load config data onto the object for logic
        pass
